/**
**********************************************************
* \file Radmor.c
* \author (c) 2023 Janusz J. Mlodzianowski
*
* Obsluga modulu Sparkfun Si4703 (bazujaca na kodzie z systemu Arduino)
* wyswietlacza SH1106 oraz odbiornika Radmor
*
* Do obslugi Si4703 wymagane sa:
* linia SDA(B0) dane interfejsu I2C
* linia SCL(B1) zegar interfejsu I2C
* linia RST(B2) reset ukladu
* Adres I2C ukladu Si4703: 0x10(B7..B1)+B0=1(read),B0=0(write)
* Rezystory pullup do I2C sa na module Si4703
*
* Do obslugi 8 przyciskow przez dzielnik napiecia 68k/10k i bufor HCT244
* Przyciski obsluguje PORTB
* B7: domyslny, skrajny z lewej, B6..B0: kolejne
*
* Do obslugi SH1106:
* linia SCK (C3)
* linia DOUT (C2)
* linia RST (C1)
* linia DC (C0)
* linia CS	(C4)
*
* Do obslugi komunikacji przez UART:
* linia RC6 (TxD) i RC7 (RxD)
************************************************************/
#define DATE "4-12-2022"
#define VERSION	"Radmor, Si4703 ver. 1.2"

#include <16f876a.h>
#include <int16cxx.h>

#pragma config|=0x3F39

//Definicje pinow do SI4703
#pragma bit SDA		@PORTA.1	//linia SDA (A1)
#pragma bit SCL		@PORTA.0	//linia SCL (A0)
#pragma bit RST		@PORTA.2	//linia RESET (A2)
#define DIR_SDA		TRISA.1		//bity TRIS
#define DIR_SCL		TRISA.0

//Definicje pinow do SH1106
#pragma bit SH1106_SCK	@PORTC.3
#pragma bit SH1106_DOUT @PORTC.2
#pragma bit SH1106_RST	@PORTC.1
#pragma bit SH1106_DC   @PORTC.0
#pragma bit SH1106_CS		@PORTC.4


extern void SH1106init(uns8 mode);
extern void SH1106clear(void);
extern uns8 SH1106gotoXY(uns8 x, uns8 y);
extern void SH1106string(const uns8 *str);
extern void SH1106int(uns8 n);

interrupt inthandler(void);
void putcharISR(char data);
char serhit(void);
char getser(void);
void usage(void);
void serinit(void);
void putchar(char data);
void putstr(const char *str);
void putuns16(uns16 num);
char eeread8(uns8 adr);
uns16 eeread16(uns8 adr);
void eewrite8(uns8 adr, char data);
void eewrite16(uns8 adr, uns16 data);
char getCommand(void);
uns16 getuns16(void);

//Bit Bang I2C do SI4703
void delay_ms(uns16 ms);
void i2cDelay(void);
void i2cInit(void);
void i2cStart(void);
void i2cStop(void);
void i2cHigh(void);
void i2cLow(void);
char i2cOut(char val);
char i2cIn(char ack);

//Si4703
void readRegisters(void);
char updateRegisters(void);
void si4703init(void);
void setVolume(char volume);
char getVolume(void);
void setChannel(uns16 channel);
uns16 getChannel(void);
uns16 seek(char dir);
uns16 seekUp(void);
uns16 seekDown(void);
char getsignal(void);

void readRDS(void);
void SH1106MHz(uns16 n);
void SH1106status(uns8 ch);

void main(void);

//Si4703 register names
#define  DEVICEID			0x00
#define  CHIPID				0x01
#define  POWERCFG			0x02
#define  CHANNEL			0x03
#define  SYSCONFIG1 	0x04
#define  SYSCONFIG2 	0x05
#define  STATUSRSSI 	0x0A
#define  READCHAN			0x0B
#define  RDSA					0x0C
#define  RDSB					0x0D
#define  RDSC					0x0E
#define  RDSD					0x0F

//Register 0x02 - POWERCFG
#define  SMUTE	15
#define  DMUTE	14
#define  SKMODE 10
#define  SEEKUP 9
#define  SEEK		8

//Register 0x03 - CHANNEL
#define  TUNE 15

//Register 0x04 - SYSCONFIG1
#define  RDS	12
#define  DE		11

//Register 0x05 - SYSCONFIG2
#define  SPACE1 5
#define  SPACE0 4

//Register 0x0A - STATUSRSSI
#define  RDSR		15
#define  STC		14
#define  SFBL		13
#define  AFCRL	12
#define  RDSS		11
#define  STEREO 8

#define  FAIL		 0
#define  SUCCESS 1

#define  SEEK_DOWN	0
#define  SEEK_UP		1

//kopia 16 (16bitowych) rejestrow SI4703
uns16 si4703_registers[16];

/**********************************************************
* Przerwania
* 4MHz:
* podstawa czasu TMR1:100ms, preskaler: 8, internal clk (1us)
* 0xffff-12500=CF2B
**********************************************************/

/**********************************************************
* struktura EEPROM'u
* 8*uns16
**********************************************************/
/* Przerwaniowa obsluga komunikacji z PC */
#pragma rambank 2

#define RXSIZE	40
char rxbuf[RXSIZE];			//bufor odbiorczy
char head,tail;
char buf[RXSIZE];
char rx;								//wskaznik do bufora
char xonxoff;						//flaga wyslania XON/XOFF
#define XON		0x11			//kody sterujace protokolem Xon/Xoff
#define XOFF	0x13

#pragma rambank 0
uns16 czas;
uns16 rdsON;														 //0 lub ms

/**
**********************************************************
* low priority interrupt handler
* Watek odczytu danych z RS232 jest uruchamiany gdy przyjdzie znak.
* Kolowy bufor odbiorczy na 40 znakw ma za zadanie odebrac jeden rozkaz
**********************************************************/
#pragma origin 4

void put16ISR(uns16 num);
interrupt inthandler(void)
{
	int_save_registers

	if(RCIE)
		if(RCIF){
			char saveFSR=FSR;
			char ch;
			char next;

		  ch=RCREG;
//			if(ch!=0x0A){ //juz tu ignorowac <lf>
				next=head+1;
				if(next>=RXSIZE) next=0;

				if(next==tail){	//bufor pelny: wyslac XOFF
					putcharISR(XOFF);
					xonxoff=0;
				}
				rxbuf[head]=ch;
				head=next;

      	putcharISR(ch);	//wyslac echo odebranego znaku

			FSR=saveFSR;
		}

	int_restore_registers
}

#include <math16.h>
#define FP_OPTIM_SIZE
#define DISABLE_ROUNDING

/**
**********************************************************
* wait and transmit char via rs232
* tylko do wysylania echo w przerwaniu
*
* @param data bajt do wyslana
**********************************************************/
void putcharISR(char data)
{
	while(TXIF==0) clrwdt();			//wait while Tx buffer empty
	TXREG=data;
}

/**
************************************************************
* buffer has NO data if and only if HEAD==TAIL
*
* @retval 1:dane sa w buforze, 0:bufor pusty
************************************************************/
char serhit(void)
{
	if(head==tail) return 0;
	return 1;
}

/**
************************************************************
* read the buffer, pull the next character from the
* buffer tail and advance it circularly
*
* @retval znak z bufora
***********************************************************/
char getser(void)
{
	char ch;
	char *next;

	if(!serhit()) return(0);     //if no data then return NULL

	next=tail+1;
	if(next>=RXSIZE) next=0;
	ch=rxbuf[tail];
	tail=next;
	if(xonxoff==0){							//w razie potrzeby wyslac XON
		putchar(XON);
		xonxoff=1;
	}
	return(ch);									 //return the character
}

#pragma codepage 3
/**
**********************************************************
* help screen
**********************************************************/
void usage(void)
{
char mode;

	putstr(VERSION" ("DATE")\r\nCopyright (C) Janusz J. Mlodzianowski, 2022\r\nfizjm@ug.edu.pl\r\n");
	putstr("Supported commands:\r\n");
	putstr("Wn,m: store to memory n channel m (n=1-8, m=MHz*10)\r\n");
	putstr("R:    read memory\r\n");
	putstr("Pn:   set current station from memory n (n=1-8)\r\n");
	putstr("Sm:   set current station to channel m (m=MHz*10)\r\n");
	putstr("G:    get current station (MHz*10)\r\n");
	putstr("Vn:   set volume to n (n=0..15)\r\n");
	putstr("+:    seek Up\r\n");
	putstr("-:    seek Down\r\n");

  putstr("Dn:   set RDS timeout to n (n=0..ms)\r\n");
  putstr("In:   set LCD background to n (n=0,1)\r\n");

	putstr("?:    this help\r\n");

	putstr("Selected channel: ");
	putuns16(getChannel());					//numer kanalu
  putstr(", Signal: ");
  putuns16(getsignal());					//wielkosc sygnalu
  putstr(" dBuV");
  if(rdsON)
  	putstr(", RDS is on");
	else
  	putstr(", RDS is off");

  mode=eeread8(18);
	if(mode)
  	putstr(", display=normal(1)\r\n>");
	else
  	putstr(", display=inverse(0)\r\n>");
}

/**
**********************************************************
* To jest dla 4MHz
* init serial port: 19200,8,NP,1
**********************************************************/
void serinit(void)
{
	char c;

	SPBRG=12;
	TXSTA=0x26;
	RCSTA=0x90;
}

/**
**********************************************************
* wait and transmit char via rs232
*
* @param data bajt do wyslana
**********************************************************/
void putchar(char data)
{
	while(TXIF==0) clrwdt();			//wait while Tx buffer empty
	TXREG=data;
}

/**
**********************************************************
* wait and transmit string via rs232
*
* @param napis do wyslana
**********************************************************/
void putstr(const char *str)
{
	while(*str) putchar(*str++);
}

/**
**********************************************************
* wait and transmit uns16 via rs232
* Cyfry wysylane od lewej do prawej
*
* @param num uns16 do wyslania
**********************************************************/
void putuns16(uns16 num)
{
	uns16 m,tmp;
	char dig,zero;

	m=10000;
	zero=0;

	if(num){
		while(m){
			dig=num/m;
			if(dig) zero=1;
			tmp=num%m;
			num=tmp;
			if(zero)putchar(dig+'0');
			m=m/10;
		}
	}else
		putchar('0');
}


/*********** Obsluga pamieci EEPROM **********************/
/**
**********************************************************
* read byte from internal PIC EEPROM
*
* @param adr adres w EEPROM
* @retval odczytany bajt
**********************************************************/
char eeread8(uns8 adr)
{
	EEADR=adr;								//address/data
  EEPGD=0;
	RD=1;											//initiate read
	return(EEDATA);
}

/**
**********************************************************
* read uns16 from internal PIC EEPROM
*
* @param adr adres w EEPROM
* @retval odczytany uns16
**********************************************************/
uns16 eeread16(uns8 adr)
{
	uns16 data;

  data.low8=eeread8(adr);
	data.high8=eeread8(adr+1);
	return(data);
}

/**
**********************************************************
* write byte to internal PIC EEPROM
*
* @param adr adres w EEPROM
* @param data bajt do wpisania
**********************************************************/
void eewrite8(uns8 adr, char data)
{
  EEPGD=0;
	WREN=1;										//enable writes
	EEADR=adr;								//address/data
	EEDATA=data;
	GIE=0;										//disable ints
	EECON2=0x55;							//magic numbers
	EECON2=0xAA;
	WR=1;											//initiate write
	GIE=1;
	while(WR==1);							//wait till done
	EEIF=0;
	WREN=0;
}

/**
**********************************************************
* write uns16 to internal PIC EEPROM
*
* @param adr adres w EEPROM
* @param data uns16 do wpisania
**********************************************************/
void eewrite16(uns8 adr, uns16 data)
{
	eewrite8(adr,  data.low8);
	eewrite8(adr+1,data.high8);
}


#pragma codepage 1

#pragma rambank 3

/**
*****************************************************************************
* Wait and return command from rs232.
* Odebrany rozkaz musi konczyc sie <cr> lub <lf>
* Odebrany rozkaz jest zapamietywany w globalnym buf a na koncu
* dodawany jest NULL (<cr>/<lf> nie sa zapamietywane)
* UWAGA:
* w trybie terminalowym nalezy uwzglednic podstawowe sekwencje VT100
* (obslugiwane przez Hyperterminal i putty) dotyczace kursora:
*
* <ESC>[A	 	up			Putty, Hyperterminal, ignorujemy
* <ESC>[B	 	down		Putty, Hyperterminal, ignorujemy
* <ESC>[C	 	right		Putty, Hyperterminal, ignorujemy
* <ESC>[D	 	left    Putty, Hyperterminal
*
* <ESC>[H		Home		Hyperterminal
* <ESC>[K		End			Hyperterminal, ignorujemy
*
* <ESC>[1~	Home		Putty
* <ESC>[2~	Insert	Putty, ignorujemy
* <ESC>[3~	Del			Putty
* <ESC>[4~	End			Putty, ignorujemy
* <ESC>[5~	PgUp		Putty, ignorujemy
* <ESC>[6~	PgDn		Putty, ignorujemy
*
* pozostale sekwencje to blad (zwraca znak 'E'
*
* @retval pierwszy znak wprowadzonego wiersza. Dla blednej sekwencji <ESC> zwraca 'E'
* oraz wypelniony buf
****************************************************************************/
char getCommand(void)
{
	char c;
  char state;
  char err;

  rx=0;
  state=0;
  err=0;

  while(1){
		c=getser();
//  	if(c==0) continue;

    if(c==0x0D) break;							//<CR> to koniec rozkazu
    if(c==0x0A) break;							//<LF> to koniec rozkazu

    if((c==0x08)||(c==0x7F)){	  		//<BS>
    	if(rx) rx--;
     	continue;
    }
    if(c==0x09) c=' ';							//<TAB> -> <SPACE>
		if(c==0x1B){										//poczatek sekwencji <ESC>?
			state=1;
  		continue;
		}


    if(c<' ') continue;							//ignorujemy pozostale znaki <ctrl>
		if(c>127) continue;							//tylko ASCII

		if(state==1){										//czy jest <ESC>[?
			if(c=='['){										//jest <ESC>[
				state=2;
				continue;
			}else
				err=1; 	//wrong
    }

		if(state==2){										//kontynuacja <ESC>[
			switch(c){
				case 'H':										//<ESC>[H		Home(Hyperterminal)
					rx=0;											//zerujemy kursor

  			case 'D': 									//<ESC>[D	 	left    (<ESC>[<num>D
   				if(rx) rx--;							//cofamy kursor
					if(rx==0)err=0;						//kasujemy blad

  			case 'A': 									//<ESC>[A	 	up			ignorujemy
  			case 'C': 									//<ESC>[C	 	right		ignorujemy
  			case 'B': 									//<ESC>[B	 	down		ignorujemy
    		case 'K':										//<ESC>[K		End(Hyperterminal) ignorujemy
					state=0;									//koniec sekwencji
      		continue;

				case '1':										//czy to poczatek <ESC>[..~ ?
    		case '2':
				case '3':
    		case '4':
				case '5':
				case '6':
    			state=0x80|c;							//zapamietac kod i sygnalizowac dalej
      		continue;

 				default:  //tu tez bez default..
	    		err=1;	//wrong
  		}
		}

		if(state>0x80){									//byc moze <ESC>[..~ (tylko putty?)
			if(c=='~'){										//jest <ESC>[..~
				switch(state&0x7F){
					case '1':									//<ESC>[1~	Home(putty)
						rx=0;                   //zerujemy kursor

    			case '3':									//<ESC>[3~	Del
						if(rx) rx--;
						if(rx==0)err=0;					//kasujemy blad
      			state=0;								//koniec sekwencji
      			continue;

					case '5':									//<ESC>[5~	PgUp	ignorujemy
          case '6':									//<ESC>[6~	PgDn	ignorujemy
    			case '4':									//<ESC>[4~	End		ignorujemy
    			case '2':									//<ESC>[2~	Insert	ignorujemy
      			state=0;                //koniec sekwencji
        		err=0;
      			continue;

    			default:                  //tu chyba mozna pozbyc sie default i else..
    				err=1;//wrong
  			}
  		}else
      	err=1;//wrong
		}

//tu jest jakby if(state==0)
		if(c>='a')c&=0xDF;							//zamienic male litery na duze
		buf[rx]=c;
    rx++;
    clrwdt();
	}

	buf[rx++]='\0';										//koniec rozkazu
	buf[rx++]='\0';

  rx=0;
  while((buf[rx]==' ')&&(buf[rx]!='\0'))rx++;		//ominac wiodace spacje
  c=rx;																					//wskazuje na kod rozkazu
  rx++;															//rx wskazuje poczatek kodu numerycznego

	if(err) return(0x1B);		 					//nie obslugiwana sekwencja <ESC>

	return(buf[c]);
}

#pragma codepage 0

#pragma rambank 0

/**
*****************************************************************************
* pobrac z bufora argument rozkazu
*
* @retval 16bitowa wartosc lub 0
****************************************************************************/
uns16 getuns16(void)
{
	uns16 val=0;
  char i=0;						//licznik znakow

//ominac spacje i przecinki
	while(((buf[rx]==' ')||((buf[rx]==','))) && (buf[rx]!='\0'))rx++;

	while((buf[rx]!='\0') && (buf[rx]>='0') && (buf[rx]<='9') ){
  	val=10*val;				//kod=10*kod+buf[rx]-48;
    val+=buf[rx];
    val-=48;
    rx++;
    i++;
  }

  if(i)
		return(val);
  else								//command jest bez kodu rozkazu
		return(0);
}



/******************** Bit Bang I2C do Si4703 **********************/
/**
****************************************************************************
* Pauza w ms przy zalozeniu zegara 4MHz
* Minimalna pauza to 1ms
*
* Pauza w ms przy zalozeniu zegara 20MHz podstawa 5MHz, preskaler 1:8 0.2*8us=1.6uS
* Minimalna pauza to 1.6us
*
* @param ms pauza w ms
***************************************************************************/
void delay_ms(uns16 ms)
{
	T1CON=0x31;       	//4MHz: internal clock, preskaler 1:8 1us*8=8uS

  do{
		TMR1H=0;					//reset TMR1
		TMR1L=0;
    clrwdt();
    while(TMR1L<125);	//4HMHZ: 125*8uS=1ms

  }while(--ms>0);
}

/**
********************************************************
* i2c delay max SCL=400kHz
***********************************************************/
void i2cDelay(void)
{
	nop();
	nop();
}

/**
************************************************************
* Inicjalizacja linii obslugujacych interfejs I2C
************************************************************/
void i2cInit(void)
{
	DIR_SDA=0;
	DIR_SCL=0;

  SDA=1;
  i2cDelay();
  SCL=1;
  i2cDelay();
}

/**
**********************************************************
* Generacja bitu start w pakiecie I2C
* SDA 100
* SCL 110
***********************************************************/
void i2cStart(void)
{
	DIR_SDA=0;		//outputs
	DIR_SCL=0;

	SDA=1;
	i2cDelay();
  SCL=1;
	i2cDelay();
	SDA=0;
	i2cDelay();
	SCL=0;
	i2cDelay();
}

/**
************************************************************
* Genaracja bitu stop w pakiecie I2C
* SDA 011
* SCL 111
************************************************************/
void i2cStop(void)
{
	SDA=0;
	i2cDelay();
	SCL=1;
	i2cDelay();
	SDA=1;
	i2cDelay();
}

/**
********************************************************
* Wystawienie na szyne I2C bitu danych o wartosci 1
* bit 1 jest wpisany roznacym zboczem SCL
* SDA 111
* SCL 010
***********************************************************/
void i2cHigh(void)
{
	SDA=1;
	i2cDelay();
	SCL=1;
	i2cDelay();
	SCL=0;
	i2cDelay();
}

/**
*********************************************************
* Wystawienie na szyne I2C biru danych i wartosci 0
* bit 0 jest wpisany roznacym zboczem SCL
* SDA 000
* SCL 010
************************************************************/
void i2cLow(void)
{
	SDA=0;
	i2cDelay();
	SCL=1;
	i2cDelay();
	SCL=0;
	i2cDelay();
}

/**
************************************************************
* Wyslanie bajtu (MSB...LSB, po czym wczytanie bitu ACK) na szyne I2C
*
* @param val bajt do wyslania
* @retval bit ACK z urzadzenia peryferyjnego
************************************************************/
char i2cOut(char val)
{
  char m;

	DIR_SDA=0;             //output

	for(m=0x80;m;m>>=1) if(val&m) i2cHigh();else i2cLow();

	SDA=1;										//read ACK bit
	DIR_SDA=1;						//input
	SCL=1;
	i2cDelay();
	m=SDA;
	SCL=0;
	i2cDelay();
	DIR_SDA=0;						//output
	i2cDelay();

  return(m);
}

/**
************************************************************
* Wczytanie bajtu (MSB..LSB, po czym wyslanie bitu ACK)
*
* @param ack bit ACK do wyslania
* @retval odczytany bajt
***********************************************************/
char i2cIn(char ack)
{
  char m,i;
  char val=0;

	SDA=1;
	DIR_SDA=1;							//input
	for(m=0x80;m;m>>=1){
		SCL=1;
		i2cDelay();
		i=SDA;
		SCL=0;
		i2cDelay();
		if(i) val|=m;
    i2cDelay();
	}
	DIR_SDA=0;							//output
	if(ack)									//ack bit (zwykle LOW)
		i2cHigh();
	else
		i2cLow();

  return(val);
}

/******************** Obsluga Si4703 **********************/

/**
************************************************************
* Odczytanie do globalnej tablicy "si4703_registers"
* stanu wszystkich 16(16 bitowych) rejestrow ukladu Si4703.
* Potwierdzenie przyjecia bajtu adresu to bit ACK=0.
* Na koncu transferu pakietu ACK=1
************************************************************/
void readRegisters(void)
{
  char x,ack;
  uns16 val;

  i2cStart();
  ack=i2cOut(0x21);   //Adres 0x10 + bit 1 oznaczajacy odczyt z

//Remember, register 0x0A comes in first so we have to shuffle the array around a bit
  for(x=0x0A; ;x++){									 	//read in these 16 registers=32 bytes
   if(x==0x10) x=0; 										//loop back to zero
  	val.high8=i2cIn(0);                 //wyslac potwierdzenie ACK=0

    if(x==0x09)          //wyslac potwierdzenie bajtu: ACK=0, na koncu pakietu: ACK=1
    	val.low8=i2cIn(1);
    else
    	val.low8=i2cIn(0);

    si4703_registers[x]=val;

    if(x==0x09) break; 									//we're done!
  }
  i2cStop();
}

/**
************************************************************
* Zapisanie z globalnej tablicy "si4703_registers"
* stanu 6 rejestrow ukladu Si4703
*
* Write the current 9 control registers (0x02 to 0x07) to the Si4703
* It's a little weird, you don't write an I2C addres
* The Si4703 assumes you are writing to 0x02 first, then increments
*************************************************************/
char updateRegisters(void)
{
	char i,ack;
  char high,low;

  i2cStart();
  ack=i2cOut(0x20);   		//Adres 0x10+bit 0 oznaczajacy zapis do

  for(i=0x02; i<0x08; i++){
    high=si4703_registers[i].high8;;   //>>8
    low=si4703_registers[i].low8;			//&0x00FF;

    ack=i2cOut(high);			 //i2cOut() zwraca odczytany bit ACK powinien byc 0
    ack=i2cOut(low);			 //Lower 8 bits
  }
  i2cStop();

//dokumentacja mowi aby po zapisie poczekac ok 500ms na stabilizacje kwarcu..
	delay_ms(500);

  return(SUCCESS); 	//a moze ack...
}

/**
************************************************************
* Inicjalizacja ukladu Si4703.
*
* To get the Si4703 inito 2-wire mode, SEN needs to be high and SDIO needs to be low after a reset
* The breakout board has SEN pulled high, but also has SDIO pulled high. Therefore, after a normal power up
* The Si4703 will be in an unknown state. RST must be controlled
*
* wymuszenie obslugi interfejsu I2C:
* linia SDA(B0) ustawiona w LOW
* ustawienie linii RESTET w LOW, pauza, ustawienie linii RESET w HIGH
* nieaktywne linie SDA i SCL: HIGH
* Rejestry po reset: 4674,4608,26385,0,6144,17,0,33024,0,..
************************************************************/
void si4703init(void)
{
	char i;

  TRISB.0=0;						//outputs
  TRISB.1=0;
  TRISB.2=0;

	SDA=0;								//aby wymusic interfejs I2C: SDA(B0) low
	SCL=1;								//SCL(B1) high

	RST=0;								//reset(B2) Si4703
  delay_ms(500);
  RST=1;
  delay_ms(500); 				//po reset pauza conajmniej 300ns
	i2cInit();						//to ustawi SDA=1, SCL=1

  readRegisters();               //Read the current register set
  si4703_registers[0x07]=0x8100; //Enable the oscillator, from AN230 page 9, rev 0.61 (works)
  updateRegisters();

  readRegisters();

  si4703_registers[POWERCFG]=0x4001; //Enable the IC
  si4703_registers[SYSCONFIG1]|=((uns16)1<<RDS); 		 //Enable RDS
  si4703_registers[SYSCONFIG1]|=((uns16)1<<DE); 		 //50kHz Europe setup
  si4703_registers[SYSCONFIG2]|=((uns16)1<<SPACE0);	 //100kHz channel spacing for Europe
  si4703_registers[SYSCONFIG2]&=0xFFF0;							 //Clear volume bits
  si4703_registers[SYSCONFIG2]|=0x0001;							 //Set volume to lowest

  updateRegisters();
}

/**
************************************************************
* Ustawienie glosnosci
*
* @param volume glosnosc w zakresie 0(min)..15(max)
************************************************************/
void setVolume(char volume)
{
  readRegisters(); //Read the current register set

  if(volume>15) volume=15;
  si4703_registers[SYSCONFIG2]&=0xFFF0; //Clear volume bits
  si4703_registers[SYSCONFIG2]|=volume; //Set new volume
  updateRegisters();
}

/**
************************************************************
* Pobranie glosnosci
*
* @retval volume glosnosc w zakresie 0(min)..15(max)
************************************************************/
char getVolume(void)
{
	uns16 vol;

  readRegisters(); //Read the current register set
  vol=si4703_registers[SYSCONFIG2];
  return(vol.low8&0x0F);
}

/**
***********************************************************
* Odczytanie z ukladu SI4703 aktualnie wybranego kanalu
*
* @retval wybrany kanal w MHz*10
************************************************************/
uns16 getChannel(void)
{
	uns16 channel;

  readRegisters();
  channel=si4703_registers[READCHAN]&0x03FF; //Mask out everything but the lower 10 bits
  channel+=875; //98+875=973

  return(channel);
}

/**
************************************************************
* Zaprogramowanie odbieranego kanalu FM
*
* @param channel numer kanalu w MHz*10
************************************************************/
void setChannel(uns16 channel)
{
  uns16 newChannel=channel*10;	//973*10=9730

  newChannel-=8750;						//9730-8750=980
  newChannel/=10;							//980/10=98

//AN230 page 20 rev 0.5
  readRegisters();
  si4703_registers[CHANNEL]&=0xFC00;			//Clear out the channel bits
  si4703_registers[CHANNEL]|=newChannel;	//Mask in the new channel
  si4703_registers[CHANNEL]|=((uns16)1<<TUNE);		//Set the TUNE bit to start
  updateRegisters();

  delay_ms(100); 				//po reset pauza conajmniej 300ns

  while(1){					//poll to see if STC is set
    readRegisters();
    if((si4703_registers[STATUSRSSI]&((uns16)1<<STC))!=0) break; //Tuning complete!
  }

  readRegisters();
  si4703_registers[CHANNEL]&=~((uns16)1<<TUNE); //Clear the tune after a tune has completed
  updateRegisters();

  while(1){					//wait for the si4703 to clear the STC as well
    readRegisters();
    if((si4703_registers[STATUSRSSI]&((uns16)1<<STC))==0) break; //Tuning complete!
  }
}

/**
************************************************************
* Szukanie nastepnej stacji
*
* @param seek kierunek szukania stacji
* @retval znaleziony kanal w MHz*10
************************************************************/
uns16 seek(char dir)
{
	uns16 valueSFBL;

  readRegisters();
//Set seek mode wrap bit
  si4703_registers[POWERCFG]|=((uns16)1<<SKMODE);		//Allow wrap
//si4703_registers[POWERCFG]&=~(1<<SKMODE); //Disallow wrap - if you disallow wrap, you may want to tune to 87.5 first
  if(dir==SEEK_DOWN)
  	si4703_registers[POWERCFG]&=~((uns16)1<<SEEKUP);	//Seek down is the default upon reset
  else
  	si4703_registers[POWERCFG]|=((uns16)1<<SEEKUP);		//Set the bit to seek up

  si4703_registers[POWERCFG]|=((uns16)1<<SEEK);			//Start seek
  updateRegisters();													//Seeking will now start

  while(1){																		//Poll to see if STC is set
    readRegisters();
    if((si4703_registers[STATUSRSSI]&((uns16)1<<STC))!=0) break; //Tuning complete!
  }

  readRegisters();
  valueSFBL=si4703_registers[STATUSRSSI]&((uns16)1<<SFBL); //Store the value of SFBL
  si4703_registers[POWERCFG]&=~((uns16)1<<SEEK); 					//Clear the seek bit after seek has completed
  updateRegisters();

  while(1){													//Wait for the si4703 to clear the STC as well
    readRegisters();
    if((si4703_registers[STATUSRSSI]&((uns16)1<<STC))==0) break; //Tuning complete!
  }

  if(valueSFBL)return(0);						//The bit was set indicating we hit a band limit or failed to find a station

  return(getChannel());
}

/**
************************************************************
* Szukanie w kierunku wyzszych czestotliwosci nastepnej stacji
*
* @retval znaleziony kanal w MHz*10
************************************************************/
uns16 seekUp(void)
{
  return(seek(SEEK_UP));
}

/**
************************************************************
* Szukanie w kierunku nizszych czestotliwosci nastepnej stacji
*
* @retval znaleziony kanal w MHz*10
************************************************************/
uns16 seekDown(void)
{
  return(seek(SEEK_DOWN));
}

/**
************************************************************
* Zwrocenie wielkosci sygnalu RSSI
*
* @retval wielkosc sygnalu w dBuV
************************************************************/
char getsignal(void)
{
	char rssi;

  readRegisters();
  rssi=si4703_registers[STATUSRSSI].low8;

  return(rssi);
}



///////////////////// obsluga RDS /////////////////////////
//RDS specific stuff
#define RDS_PS (0)
#define RDS_RT (2)

#pragma rambank 3
char rdstmp[8];
char rdsdata[22];
uns16 pscharindex;
const rs[]="---JJM, 4-12-2022---\0";    //domyslny RDS (21 znakow)


/**
************************************************************
* Pobranie z sygnalu RDS pakietu PS (8 znakow)
*
* \note funkcja powinna byc wywolywana cyklicznie lub jako
* odpowiedz na przerwanie.
* W praktyce pakiet PS nie zawiera tekstu statycznego a ciagle
* zmieniajacy sie...
* \note odczyt RDS powoduje szum wiec w praktyce RDS jest wlaczony
************************************************************/
void readRDS(void)
{
  char Dh,Dl;
  char Ch,Cl;
  uns16 b,c,d;
  char i,cc;

	uns16 grouptype;
  uns16 receivedcharindex;

  static uns8 sig;

  if(rdsON==0) return; 	//obsluga wylaczona

	if(czas){
		czas--;
		return;
	}

  readRegisters();
	if(si4703_registers[STATUSRSSI]&0x8000){		//jest wiadomosc RDS
		b=si4703_registers[RDSB];
	  c=si4703_registers[RDSC];
	  d=si4703_registers[RDSD];

    grouptype=b>>11;
		switch(grouptype){
			case 0:										//pakiet PS
      case 1:
    	  receivedcharindex=b&0x03;
        receivedcharindex<<=1;
        if(pscharindex==receivedcharindex){
  				Dh=d.high8;
    			Dl=d.low8;
					rdstmp[receivedcharindex]=Dh;
					rdstmp[receivedcharindex+1]=Dl;
					pscharindex+=2;

					if(pscharindex==8){					//czy nowy tekst?
						for(i=0;i<8;i++){
							cc=rdsdata[i+8];
  						if(rdstmp[i]!=cc)break;	//nowy i stary sa rozne
            }

						if(i<8){									//nowy tekst
							for(i=0;i<8;i++){				//stary przesunac o 8 w lewo
								cc=rdsdata[i+8];
  							rdsdata[i]=cc;
							}
							for(i=0;i<8;i++){				//wpisac nowy
               	cc=rdstmp[i];
  							rdsdata[i+8]=cc;
              }
						}
            rdsdata[16]='\0';					//wydruk RDS
            SH1106gotoXY(2*6,0);
            SH1106string(rdsdata);
            pscharindex=0;

						if(sig!=si4703_registers[STATUSRSSI].low8){			//update mocy sygnalu
							sig=si4703_registers[STATUSRSSI].low8;
							SH1106gotoXY(7*6,7);
							SH1106int(sig);
							SH1106string("dBuV");
						}
          }
					czas=rdsON;
        }else
         	pscharindex=0;
          break;
		}
	}

	delay_ms(30);
}
#pragma rambank 0

#include "sh1106SPI.c"

#pragma rambank 1
#pragma codepage 3

/**
**********************************************************************
* wyswietlenie (od prawej do lewej..) czcionka 32x16 czestotliwosci w MHz
*********************************************************************/
void SH1106MHz(uns16 n)
{
  char buf[6];
  char *s=buf+5;
  char c,i;

  *s='\0';
  i=1;
  do{
    c=n%10;
    n/=10;
    if(i==2)*--s='.';
    *--s=c+'0';    //c<10?(c+'0'):(c+'A'-10);
    i++;
  }while(n);

  SH1106clear();

//  SH1106gotoXY( (128-(i*17+21))/2, 2);  //cyfr*17pikseli+MHz(21pikseli)
	c=SH1106_XPIXELS-(i*17+21);
  c/=2;
 	SH1106gotoXY(c, 2);

  SH1106string3216(s);

  SH1106gotoXY(SH1106column,5);
  SH1106string(" MHz");
}

/**
**********************************************************************
* wyswietlenie statusu
*********************************************************************/
void SH1106status(uns8 ch)
{
  uns8 sig,vol;

	readRegisters();

  sig=si4703_registers[STATUSRSSI].low8;
  vol=si4703_registers[SYSCONFIG2].low8;
  vol&=0x0F;

	SH1106gotoXY(0,7);
  SH1106string("ch:");
  SH1106int(ch+1);

	SH1106gotoXY(7*6,7);
  SH1106int(sig);
  SH1106string("dBuV");

	SH1106gotoXY(15*6,7);
  SH1106string("vol:");
  SH1106int(vol);
}

/**
**********************************************************************
* przekopiowanie domyslnego RDS
*********************************************************************/
void strcpy(void)
{
	char i;

	for(i=0;i<22;i++) rdsdata[i]=rs[i];	//strcpy
}

#pragma codepage 0
#pragma rambank 0

/**
********************************************************
* main
*
* Odebrany z PC rozkaz konczy sie <cr> lub <lf> po czym jest wysylany znak prompt '>'
* Wyslana do PC odpowiedz konczy sie <cr><lf>
* Obslugiwane rozkazy:
* Wn,m<cr>
* Zaprogramowac w EEPROM n stacje m (n=1-8, m=MHz*10)
* Odpowiedz: "Set memory n to m" lub komunikat bledu
*
* R<cr>
* Odczytac zaprogramowane w EEPROM stacje m
* Odpowiedz: "Channels: m1, m2, m3, m4, m5, m6, m7, m8"
*
* Pn<cr>  (n=1-8)
* Ustawic odbior na zaprogramowany w EEPROM kanal n
* Odpowiedz: "Set channel n, Signal x dBuV" lub komunikat bledu
*
* Sm<cr>  (m=MHz*10)
* Ustawic stacje na podany w rozkazie kanal m
* Odpowiedz: "Set channel m, Signal x dBuV" lub komunikat bledu
*
* G<cr>
* Odczytac aktualnie wybrany kanal
* Odpowiedz: "Selected channel m, Signal x dBuV" lub komunikat bledu
*
* Vx<cr><lf>	(x=0-15)
* Ustawic glosnosc
* Odpowiedz: "Set volume to x" lub komunikat bledu
*
* +<cr>
* SeekUP
* Odpowiedz: "Seeked to m, Signal x dBuV" lub komunikat bledu
*
* -<cr>
* SeekDown
* Odpowiedz: "Seeked to m, Signal x dBuV" lub komunikat bledu
*
* ?<cr>
* Test polaczenia
* Odpowiedz: help screen
*
* Dla nieobslugiwanej sekwencji
* Not supported <ESC> sequnce"
*
* Dla nieobslugiwanego rozkazu:
*	Not supported command..
**********************************************************/
#pragma codepage 0
#pragma rambank 0
void main(void)
{
	char cmd;
  char OldKanal;				//flaga aktualnie wybranego kanalu
  uns16 val;
  char sensor;					//numer aktywnego sensora

	PORTA=0x01;						//RST=0,SCL=1,SDA=0: jak do konfiguracji Si4703
	PORTB=0xFF;           //ale dlaczego tu musi byc 0xff..
  PORTC=0x12;						//CS=1

  TRISA=0xF8;  					//1111000
	TRISB=0xFF;						//all inputs
	TRISC=0xA0;						//C0,1,2,3,4 outputs, RX(C7=1),TX(C6=0))

//watchdog timeout ok 18ms bez prescalera
	OPTION_REG=0x8F;			//prescaler 1:128
	PIE1=0;								//clear all individual intrs
	ADCON0=0x00;					//ADC: off
	ADCON1=0x06;					//ANx: digital
	CMCON=0x07;						//RA3..RA0 digital
	CVRCON=0;							//comparators off
	EECON1=0;
	INTCON=0x00;					//wylaczyc przerwania
	PIE2=0;								//disable EEPROM interrupt,enable CCP2
	RCIE=1;								//enable UART receive intr. (PIE1|=0x20)

//dla 4MHz
//	TMR1L=0x2B;					// (1e-6)*8*65535=0.52428sek
//	TMR1H=0xCF;					// (1e-6)*8*62500=0.5sek
//	T1CON=0;//0x31;			// (1e-6)*8*12500=0.1sek
//	T1CON=0x35;

//TMR1 obsluguje licznik czasu
//	czas=1000;					//ok 1 sek
//	TMR1L=0x30;					// reset TMR1,kolejnosc
//	TMR1H=0xFC;
//	T1CON=1;						//enable TMR1
//	TMR1IE=1;						//enable TMR1 1ms inter. (PIE1|=0x01);

	cmd=eeread8(18);
	SH1106init(cmd);
  SH1106clear();

	si4703init();						//init odbiornik
	OldKanal=0x80;					//Radmor po reset wybiera pierwszy (skrajny z lewej) kanal
  sensor=0;								//numer sensora
  val=eeread16(0);				//domyslna stacja Trojka (999)
	setChannel(val);
  setVolume(15);					//domyslna glosnosc (15)

	serinit();				//konfiguracja UART

	INTCON=0;					//GIE|PEIE  disable(0), enable(0xC0)
	delay_ms(1000);		//strategicznie wazna!
	TXREG=0; 					//aby wyzerowac TXIF
	cmd=RCREG;
	cmd=RCREG;
	head=0;
	tail=0;
	xonxoff=1;
//wlaczyc
	INTCON=0xC0;

	pscharindex=0;
  strcpy();

  SH1106gotoXY(0,0);
  SH1106string(rdsdata);
	delay_ms(2000);

	SH1106clear();
	SH1106MHz(val);
	SH1106gotoXY(0,0);
	SH1106string(rdsdata);
  SH1106status(0);

	rdsON=eeread16(16);	//timeout rds
  czas=rdsON;

	rdsON=0;	//brutalnie wylaczam odczyr RDS
	czas=0;

	while(1){										//do it forever
		readRDS();

 		cmd=PORTB;
		if(OldKanal!=cmd){					//sprawdzic przyciski z PORTB
     	delay_ms(200);						//na wszelki wypadek
			cmd=PORTB;								//zwroci 128,64,32,16,8,4,2,1
			OldKanal=cmd;

			switch(cmd){
				case 0x80: cmd=0;break;		//adres kanalu w EEPROM  kolejnosc sensorow
 				case 0x40: cmd=2;break;	//lewy..prawy
 				case 0x20: cmd=4;break;
 				case 0x10: cmd=6;break;
 				case 0x08: cmd=8;break;
 				case 0x04: cmd=10;break;
       	case 0x02: cmd=12;break;
 				case 0x01: cmd=14;break;
			}
  		val=eeread16(cmd);
 			setChannel(val);					//przestawic stacje
			SH1106MHz(val);
      sensor=cmd>>1;

			SH1106gotoXY(0,0);
			SH1106string(rdsdata);
			SH1106status(sensor);
		}

		if(serhit()){								//jesli cos przyszlo
			cmd=getCommand();					//litera kodu rozkazu + wypelniony buf
			switch(cmd){
				case 'W':								//Wn,m Zaprogramowac stacje
          val=getuns16();				//numer stacji n=1=8
          if((val<1)||(val>8)){	//kontrola bledu!
          	putstr("\r\nSupported Wn,m n-values: 1-8, command ignored..\r\n>");
            break;
					}

          cmd=val.low8;
          val=getuns16();				//numer kanalu m (MHz*10);
					if(val==0){
          	putstr("\r\nMissing Wn,m m-value, command ignored..\r\n>");
            break;
          }

					putstr("\r\nSet memory: ");
					putuns16(cmd);

					cmd--;								//adres kanalu w EEPROM
					cmd<<=1;
					eewrite16(cmd,val);

					putstr(" to: ");
					putuns16(val);
					putstr("\r\n>");
          break;

				case 'R':								//R Odczytac zaprogramowane 8 stacji
					putstr("\r\nChannels: ");
          for(cmd=0;cmd<16;cmd+=2){
						val=eeread16(cmd);
            putuns16(val);
            if(cmd<14)putstr(", ");
          }
          putstr("\r\n>");
          break;

				case 'P':								//Pn Ustawic odbior na zaprogramowany wczesniej kanal
					val=getuns16();				//odczytac # kanalu 1,2,..8
          if((val<1)||(val>8)){	//kontrola bledu!
          	putstr("\r\nSupported Pn n-values: 1-8, command ignored..\r\n>");
            break;
					}

					val--;								//adres kanalu w EEPROM: 0,2,4,6,8,10,12,14
          val<<=1;
					val=eeread16(val.low8);
PP:	 			setChannel(val);
					SH1106MHz(val);
					SH1106gotoXY(0,0);
					SH1106string(rdsdata);
          SH1106status(sensor);
          putstr("\r\nSet channel: ");
          putuns16(val);
          putstr(", Signal: ");
          putuns16(getsignal());
          putstr("dBuV\r\n>");
          break;

				case 'S':								//Sm Ustawic kanal na podana w rozkazie wartosc MHz
					val=getuns16();
					if(val==0){
          	putstr("\r\nMissing Sm m-value, command ignored..\r\n>");
            break;
          }
//sprawdzic dozwolony zakres..
					goto PP;
          break;

				case 'G':								//G Odczytac aktualnie wybrany kanal MHz
					val=getChannel();
          putstr("\r\nSelected channel: ");
          putuns16(val);
          putstr(", Signal: ");
          putuns16(getsignal());
          putstr("dBuV\r\n>");
          break;

				case 'V':								//Vx Ustawic glosnosc 0..15
					val=getuns16();
          if(val>15)
          	putstr("\r\nSupported Vn n-values: 0-15, set V15\r\n>");
					else{
          	putstr("\r\nSet volume to: ");
            putuns16(val.low8);
            putstr("\r\n>");
          }
          setVolume(val.low8);
					SH1106gotoXY(0,0);
					SH1106string(rdsdata);
					SH1106status(sensor);
          break;

				case '+':								//+ seek do nastepnej stacji
					val=seekUp();
seeked:		putstr("\r\nSeeked to: ");
          putuns16(val);
          putstr(", Signal: ");
          putuns16(getsignal());
          putstr(" dBuV\r\n>");
          SH1106MHz(val);
					SH1106gotoXY(0,0);
					SH1106string(rdsdata);
          SH1106status(sensor);
          break;

				case '-':								//- seek do poprzedniej stacji
					val=seekDown();
					goto seeked;
          break;

				case '?':								//? usage
        	usage();
          break;

				case '\0':							//pusta linia..
          putstr("\r\n>");
          break;

				case 'D':
					rdsON=getuns16();				//0:off else ms
					eewrite8(16,rdsON); //pierwsza wolna po kanalach
  				czas=rdsON;

  				strcpy();
					SH1106gotoXY(0,0);
					SH1106string(rdsdata);

  				putstr("\r\n>");
					break;

				case 'I':
					val=getuns16();				//0:off else ms
					eewrite8(18,val.low8); //pierwsza wolna po rdsON

  				strcpy();
					SH1106init(val.low8);
  				SH1106clear();
  				OldKanal=0;
  				putstr("\r\n>");
					break;

        case 0x1B:
					putstr("\r\nNot supported <ESC> sequnce\r\n>");
					break;

				default:
					putstr("\r\nNot supported command..\r\n>");
			}			 //koniec switch(cmd)...
    }				 //koniec if(chkchar())...
    clrwdt();
	}					 //koniec glownej petli
}

#define CDATA_START 0x2100	//EEPROM address
#pragma cdata[CDATA_START]	//domyslne wartosci 8 kanalow jako word
#pragma cdata[]=0xE7,0x03		//Trojka					999
#pragma cdata[]=0xD8,0x03		//RMF FM					984
#pragma cdata[]=0xDB,0x03		//TokFM						987
#pragma cdata[]=0x1A,0x04		//Radio ZET				1050
#pragma cdata[]=0x74,0x03		//RMF Classic			884
#pragma cdata[]=0x06,0x04		//Zlote Przeboje 	1030
#pragma cdata[]=0xC4,0x03		//RMF MAXX				964
#pragma cdata[]=0xF9,0x03		//Plus Gdansk			1017
#pragma cdata[]=0x00,0x00		//RDS is off
#pragma cdata[]=0x00				//invers LCD backround

